import yaml
from satcfdi import Code


class NoAliasDumper(yaml.SafeDumper):
    def ignore_aliases(self, data):
        return True


def print_yaml(data):
    print(
        yaml.dump(data, Dumper=NoAliasDumper, allow_unicode=True, width=1280, sort_keys=False)
    )


def log_email(receptor, notify_invoices, facturas_pendientes):
    print_yaml({
        "rfc": Code(receptor["Rfc"], receptor["RazonSocial"]),
        "facturas": [f"{i.name} - {i.uuid}" for i in notify_invoices],
        "pendientes_meses_anteriores": [f"{i.name} - {i.uuid}" for i in facturas_pendientes],
        "correos": receptor["Email"]
    })


def header_line(text):
    ln = (151 - len(text)) // 2
    return ("=" * ln) + " " + text + " " + ("=" * ln)


def log_line(text):
    print(
        header_line(text)
    )


def log_item(text):
    ln = (151 - len(text)) // 2
    print(
        ("*" * ln) + " " + text + " " + ("*" * ln),
    )


def cfdi_header(cfdi):
    receptor = Code(cfdi['Receptor']['Rfc'], cfdi['Receptor']['Nombre'])
    return f"{cfdi.name} - {cfdi.uuid} {receptor}"



import base64
import io
import itertools
import logging
import os
from datetime import timedelta, date
from zipfile import ZipFile

from PySimpleGUI import POPUP_BUTTONS_OK_CANCEL, PySimpleGUI
from satcfdi import DatePeriod
from satcfdi.accounting import EmailManager
from satcfdi.exceptions import ResponseError, DocumentNotFoundError
from satcfdi.pacs import Accept
from satcfdi.pacs.sat import SAT, TipoDescargaMasivaTerceros, EstadoSolicitud
from satcfdi.printer import Representable
# noinspection PyUnresolvedReferences
from satcfdi.transform.catalog import CATALOGS
from tabulate import tabulate

from . import __version__, PPD, PUE
from .client_validation import validar_client
from .file_data_managers import ClientsManager, FacturasManager
from .gui_functions import generate_ingresos, pago_factura, exportar_facturas, facturas_filename, \
    periodo_desc, generate_html_template, mf_pago_fmt, print_invoices, print_cfdis, print_cfdi_details, ajustes, ajustes_directory
from .layout import make_layout, ActionButtonManager
from .localdb import LocalDBSatCFDI, LiquidatedState
from .log_tools import log_line, log_item, cfdi_header, header_line, print_yaml
from .mycfdi import get_all_cfdi, MyCFDI, move_to_folder
from .utils import random_string, to_uuid, parse_date_period, parse_ym_date, load_certificate, to_int

logging.getLogger("weasyprint").setLevel(logging.ERROR)
logging.getLogger("fontTools").setLevel(logging.ERROR)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

ACTION_FACTURAS = "facturas"
ACTION_EMAILS = "emails"
ACTION_CLIENTS = "clients"
ACTION_AJUSTES = "ajustes"


def open_launch_window():
    layout = [[PySimpleGUI.Text("New Window", key="new")]]
    window = PySimpleGUI.Window("Launch Window", layout, modal=True, size=(300, 300))
    window.read(timeout=1000)

    return window


class FacturacionGUI:
    def __init__(self, config):
        self.email_manager = EmailManager(
            **config['email']
        )
        pac = config['pac']
        pac_module, pac_class = pac['type'].split(".")
        mod = __import__(f"satcfdi.pacs.{pac_module}", fromlist=[pac_class])
        self.pac_service = getattr(mod, pac_class)(
            **pac['args']
        )
        self.serie = config['serie']
        self.csd_signer = load_certificate(config.get('csd')) if 'csd' in config else None
        self.fiel_signer = load_certificate(config.get('fiel')) if 'fiel' in config else None

        self.sat_service = SAT(signer=self.fiel_signer)
        self.rfc_prediales = config['rfc_prediales']

        self.window = PySimpleGUI.Window(
            f"Facturación Masiva CFDI 4.0 {self.csd_signer.rfc}",
            make_layout(bool(self.fiel_signer)),
            size=(1280, 800),
            resizable=True,
            font=("Courier New", 10, "bold"),
        )

        self.all_invoices = None
        self.local_db = LocalDBSatCFDI(
            enviar_a_partir=config['enviar_a_partir'],
            pagar_a_partir=config['pagar_a_partir']
        )
        # noinspection PyTypeChecker
        self.selected_satcfdi = None  # type: MyCFDI

        self.action_button_manager = ActionButtonManager(self.window["crear_facturas"])
        self.console = self.window["console"]

        MyCFDI.local_db = self.local_db

    def run(self):
        self.window.finalize()

        self.set_folio()
        self.window['serie'].update(self.serie)

        self.window['factura_pagar'].bind("<Return>", "_enter")
        self.window['periodo'].bind("<Return>", "_enter")
        self.window['importe_pago'].bind("<Return>", "_enter")
        self.window['fecha_pago'].bind("<Return>", "_enter")
        self.window['inicio'].bind("<Return>", "_enter")
        self.window['final'].bind("<Return>", "_enter")
        self.window['forma_pago'].bind("<Return>", "_enter")
        # self.window['console'].bind('<Button-3>', '_double_click')

        # Add logging to the window
        h = logging.StreamHandler(self.window['console'])
        h.setLevel(logging.INFO)
        logging.root.addHandler(h)

        self.main_loop()
        self.window.close()

    def initial_screen(self, emisor_cif):
        log_line("Acerca De")
        print_yaml({
            "version": __version__.__version__,
            "facturacion": "CFDI 4.0",
            "emisor": emisor_cif,
            "pac_service": {
                "Type": type(self.pac_service).__name__,
                "Rfc": self.pac_service.RFC,
                "Environment": str(self.pac_service.environment)
            }
        })

    def get_all_invoices(self):
        if self.all_invoices:
            return self.all_invoices
        self.all_invoices = get_all_cfdi()
        return self.all_invoices

    def generate_invoice(self, invoice):
        ref_id = random_string()
        attempts = 3
        for i in range(attempts):
            if i:
                print(f'Intentando de nuevo... Intento {i + 1} de {attempts}')
                self._read(timeout=2000 * i)

            try:
                res = self.pac_service.stamp(
                    cfdi=invoice,
                    accept=Accept.XML_PDF,
                    ref_id=ref_id
                )
            except Exception as ex:
                logger.error(f"Error Generando: {invoice.get('Serie')}{invoice.get('Folio')} {invoice['Receptor']['Rfc']}")
                if isinstance(ex, ResponseError):
                    logger.error(f"Status Code: {ex.response.status_code}")
                    logger.error(f"Response: {ex.response.text}")
                continue

            self.set_folio(int(res['Folio']) + 1)
            return move_to_folder(res.xml, pdf_data=res.pdf)

    def set_folio(self, folio: int = None):
        if folio:
            self.local_db.folio_set(folio)
            self.window['folio'].update(folio)
        else:
            self.window['folio'].update(self.local_db.folio())

    def enviar_correos(self, emisor_cif, emails):
        with self.email_manager.sender as s:
            for receptor, notify_invoices, pendientes_meses_anteriores in emails:
                attachments = []
                for r in notify_invoices:
                    attachments += [r.filename + ".xml", r.filename + ".pdf"]

                subject = f"Comprobantes Fiscales {receptor['RazonSocial']} - {receptor['Rfc']}"
                s.send_email(
                    subject=subject,
                    to_addrs=receptor["Email"],
                    html=generate_html_template(
                        'mail_facturas_template.html',
                        fields={
                            "facturas": notify_invoices,
                            'pendientes_meses_anteriores': pendientes_meses_anteriores,
                            'emisor': emisor_cif,
                        },
                    ),
                    file_attachments=attachments
                )
                for r in notify_invoices:
                    self.local_db.notified_set(r.uuid, True)
                print_yaml({
                    "correo": subject,
                    "para": receptor["Email"]
                })
                self._read()

    def recupera_comprobantes(self, id_solicitud):
        response = self.sat_service.recover_comprobante_status(
            id_solicitud=id_solicitud
        )
        print_yaml(response)
        self._read()
        if response["EstadoSolicitud"] == EstadoSolicitud.Terminada:
            for id_paquete in response['IdsPaquetes']:
                response, paquete = self.sat_service.recover_comprobante_download(
                    id_paquete=id_paquete
                )
                print_yaml(response)
                self._read()
                yield id_paquete, base64.b64decode(paquete) if paquete else None

    def unzip_cfdi(self, file):
        with ZipFile(file, "r") as zf:
            for fileinfo in zf.infolist():
                xml_data = zf.read(fileinfo)
                move_to_folder(xml_data, pdf_data=None)
                self._read()

    def _read(self, timeout=0):
        event, values = self.window.read(timeout=timeout)
        if event in ("Exit", PySimpleGUI.WIN_CLOSED):
            exit(0)

    def action_button(self, action_name, action_items):
        if action_name == ACTION_FACTURAS:
            self.all_invoices = None
            for invoice in action_items:
                cfdi = self.generate_invoice(invoice=invoice)
                if cfdi is None:
                    break
                print_yaml({
                    "FacturaGenerada": cfdi_header(cfdi),
                })
                self._read()
        elif action_name == ACTION_EMAILS:
            clients = ClientsManager()
            emisor_cif = clients[self.csd_signer.rfc]
            self.enviar_correos(emisor_cif, action_items)
        elif action_name == ACTION_CLIENTS:
            for rfc, details in action_items.items():
                print(f"Validando: {rfc}")
                self._read()
                validar_client(rfc, details)

    def set_selected_satcfdi(self, factura):
        i = factura
        self.selected_satcfdi = i
        self.window["ver_factura"].update(disabled=not i)
        if i:
            estado = self.local_db.status_sat(i).get('Estado', 'Vigente')
            self.window["status_sat"].update(
                estado,
                visible=True,
                button_color="red4" if estado != "Vigente" else "green",
            )
        else:
            self.window["status_sat"].update(
                visible=False
            )

        # Email
        is_enviable = i \
                      and i["Emisor"]["Rfc"] == self.csd_signer.rfc \
                      and i["Fecha"] >= self.local_db.enviar_a_partir \
                      and i.estatus == "1"
        if is_enviable:
            self.window["email_notificada"].update(
                " Enviada  " if self.local_db.notified(i) else "Por Enviar",
                visible=True,
                button_color="green" if self.local_db.notified(i) else "red4",
            )
        else:
            self.window["email_notificada"].update("", visible=False)

        # Pendiente de Pago
        is_pendientable = i \
                          and i["Emisor"]["Rfc"] == self.csd_signer.rfc \
                          and i["TipoDeComprobante"] == "I" \
                          and i["Fecha"] >= self.local_db.pagar_a_partir[i['MetodoPago']] \
                          and i.estatus == "1" \
                          and (i["MetodoPago"] == PUE or i.saldo_pendiente) \
                          and i["Total"]
        if is_pendientable:
            self.window["pendiente_pago"].update(
                (" Pagada  " if i["MetodoPago"] == PUE else "Ignorada ") if self.local_db.liquidated(i) else "Por Pagar",
                visible=True,
                button_color="green" if self.local_db.liquidated(i) else "red4",
            )
        else:
            self.window["pendiente_pago"].update("", visible=False)

        # PPD
        is_ppd_active = i \
                        and i["Emisor"]["Rfc"] == self.csd_signer.rfc \
                        and i.get("MetodoPago") == PPD \
                        and i.estatus == "1" \
                        and i.saldo_pendiente
        self.window["prepare_pago"].update(disabled=not is_ppd_active)
        self.window["fecha_pago_select"].update(disabled=not is_ppd_active)
        self.window["fecha_pago"].update(disabled=not is_ppd_active)
        self.window["importe_pago"].update(disabled=not is_ppd_active)
        self.window["forma_pago"].update(disabled=not is_ppd_active)

    def print_satcfdis(self, cfdis):
        def info_fmt(i):
            return "" if self.local_db.notified(i) else "📧"

        if cfdis := sorted(cfdis, key=lambda i: (i["Fecha"], i.name), reverse=True):
            if self.window['detallado'].get():
                for i, cfdi in enumerate(cfdis, start=1):
                    log_item(f"FACTURA NUMERO: {i}")
                    print_yaml(cfdi)
                    self.local_db.describe(cfdi)
            else:
                print_invoices(
                    [
                        [
                            i,
                            cfdi['Receptor'].get('Nombre', '')[0:36],
                            cfdi['Receptor']['Rfc'],
                            cfdi.name,
                            cfdi["Fecha"].strftime("%Y-%m-%d"),
                            cfdi["Total"],
                            self.local_db.liquidated_state(cfdi),
                            mf_pago_fmt(cfdi),
                            cfdi.uuid,
                            info_fmt(cfdi)
                        ]
                        for i, cfdi in enumerate(cfdis, start=1)
                    ]
                )
            if len(cfdis) == 1:
                print_cfdi_details(cfdis[0])
                self.set_selected_satcfdi(cfdis[0])
            else:
                self.set_selected_satcfdi(None)
        else:
            print("No hay resultados")
            self.set_selected_satcfdi(None)

    def print_prepared_cfdis(self, cfdis, start=1):
        if cfdis:
            if self.window['detallado'].get():
                for i, cfdi in enumerate(cfdis, start=start):
                    log_item(f"FACTURA NUMERO: {i}")
                    print_yaml(cfdi)
            else:
                print_cfdis(cfdis, start=start)
            self.action_button_manager.set_items(ACTION_FACTURAS, cfdis)
        else:
            print("No hay facturas para este mes")

    def header(self, name, clear=True):
        if clear:
            self.console.update("")
        log_line(name)

    def main_loop(self):
        while True:
            event, values = self.window.read()
            try:
                if event in ("Exit", PySimpleGUI.WIN_CLOSED):
                    return

                action_name, action_items = self.action_button_manager.clear()

                if event in ("prepare_correos", "prepare_clientes", "prepare_facturas", "crear_facturas",
                             "inicio_enter", "final_enter", "preparar_ajuste_anual", "recuperar_emitidas", "recuperar_recibidas"):
                    self.set_selected_satcfdi([])

                match event:
                    case "folio":
                        self.set_folio(to_int(values["folio"]))

                    case "about":
                        clients = ClientsManager()
                        self.initial_screen(clients[self.csd_signer.rfc])

                    case "buscar_factura" | 'factura_pagar_enter':
                        self.header("Buscar Factura")

                        if search_text := values["factura_pagar"].strip().upper():
                            if len(search_text) < 3:
                                self.console.update("El texto de búsqueda debe tener al menos 3 caracteres")
                                continue

                            if search_uuid := to_uuid(search_text):
                                def fac_iter():
                                    if v := self.get_all_invoices().get(search_uuid):
                                        yield v
                            else:
                                def fac_iter():
                                    for i in self.get_all_invoices().values():
                                        if i["Emisor"]["Rfc"] == self.csd_signer.rfc and \
                                                (i.name == search_text or i["Receptor"]["Rfc"] == search_text or search_text in i["Receptor"].get("Nombre", "")):
                                            yield i

                            self.print_satcfdis(fac_iter())
                            self.window["descarga"].update(disabled=not search_uuid)

                    case "preparar_ajuste_anual":
                        self.header(f"AJUSTES")
                        ajustes(
                            emisor_rfc=self.csd_signer.rfc,
                            ym_date=parse_ym_date(values['periodo'])
                        )

                    case "recuperar_emitidas" | "recuperar_recibidas":
                        self.header("RECUPERAR")
                        fecha_final = date.today()
                        fecha_inicial = fecha_final - timedelta(days=int(values["recuperar_dias"]))
                        id_solicitud = self.local_db.get(event)

                        if not id_solicitud:
                            self._read()
                            response = self.sat_service.recover_comprobante_request(
                                fecha_inicial=fecha_inicial,
                                fecha_final=fecha_final,
                                rfc_receptor=self.sat_service.signer.rfc if "recuperar_recibidas" == event else None,
                                rfc_emisor=self.sat_service.signer.rfc if "recuperar_emitidas" == event else None,
                                tipo_solicitud=TipoDescargaMasivaTerceros.CFDI,
                            )
                            print_yaml(response)
                            self.local_db[event] = response['IdSolicitud']
                            print("Nueva Solicitud Creada")
                        else:
                            print_yaml({
                                'IdSolicitud': id_solicitud
                            })
                            self._read()
                            for paquete_id, data in self.recupera_comprobantes(id_solicitud):
                                if data:
                                    self.all_invoices = None
                                    with io.BytesIO(data) as b:
                                        self.unzip_cfdi(b)
                                del self.local_db[event]
                            log_line("FIN")

                    case "prepare_clientes":
                        self.header("CLIENTES")
                        clients = ClientsManager()
                        ym_date = parse_ym_date(values['periodo'])
                        if clients:
                            facturas = FacturasManager(ym_date)["Facturas"]
                            print(
                                tabulate(
                                    [
                                        [
                                            i,
                                            client["RazonSocial"][:36],
                                            client["Rfc"],
                                            client["RegimenFiscal"].code,
                                            client["CodigoPostal"],
                                            sum(1 for f in facturas if f["Receptor"] == k) or None,
                                            "; ".join(client["Email"])[0:72],
                                        ]
                                        for i, (k, client) in enumerate(clients.items(), start=1)
                                    ],
                                    headers=(
                                        "",
                                        "Razon Social",
                                        "Rfc",
                                        "Reg",
                                        "CP",
                                        "#",
                                        "Email",
                                    ),
                                    colalign=("right", "left", "left", "left", "left", "right"),
                                )
                            )
                            self.action_button_manager.set_items(ACTION_CLIENTS, clients)
                        else:
                            print("No hay clientes")

                    case "prepare_facturas" | "inicio_enter" | "final_enter":
                        ym_date = parse_ym_date(values["periodo"])
                        self.header(f"PREPARAR FACTURAS {values['periodo']}")
                        print('Periodo:', periodo_desc(ym_date, 'Mensual.1'), '[AL ...]')
                        inicio = int(values["inicio"])

                        if cfdis := generate_ingresos(
                            folio=int(values["folio"]),
                            serie=self.serie,
                            clients=ClientsManager(),
                            facturas=FacturasManager(ym_date)["Facturas"],
                            ym_date=ym_date,
                            csd_signer=self.csd_signer
                        ):
                            final = to_int(values["final"]) or len(cfdis)
                            cfdis = cfdis[max(inicio - 1, 0):max(final, 0)]

                        self.print_prepared_cfdis(cfdis, start=inicio)

                    case "prepare_pago" | "importe_pago_enter" | "fecha_pago_enter" | "forma_pago_enter":
                        self.header("COMPROBANTE PAGO")
                        if i := self.selected_satcfdi:
                            if cfdi := pago_factura(
                                    serie=self.serie,
                                    folio=int(values["folio"]),
                                    factura_pagar=i,
                                    fecha_pago=values["fecha_pago"],
                                    forma_pago=values["forma_pago"],
                                    importe_pago=values["importe_pago"],
                                    csd_signer=self.csd_signer
                            ):
                                self.print_prepared_cfdis([cfdi])
                                print_cfdi_details(cfdi)

                    case "ver_factura":
                        if i := self.selected_satcfdi:
                            os.startfile(
                                os.path.abspath(i.filename + ".pdf")
                            )

                    case "status_sat":
                        self.header("STATUS SAT")
                        if i := self.selected_satcfdi:
                            estado = self.local_db.status_sat(i, update=True)
                            self.print_satcfdis([i])
                            print_yaml(estado)

                    case "pendiente_pago":
                        if i := self.selected_satcfdi:
                            st = self.local_db.liquidated_flip(i)
                            self.console.update("")
                            self.header(self.window[event].ButtonText.upper())
                            self.print_satcfdis([i])
                            if i["MetodoPago"] == PUE:
                                print(f"FACTURA MARCADA COMO {'' if st else '-NO- '}PAGADA")
                            else:
                                print(f"FACTURA MARCADA COMO {'' if st else '-NO- '}IGNORADA")

                    case "email_notificada":
                        if i := self.selected_satcfdi:
                            st = self.local_db.notified_flip(i)
                            self.console.update("")
                            self.header(self.window[event].ButtonText.upper())
                            self.print_satcfdis([i])
                            print(f"FACTURA MARCADA COMO {'' if st else '-NO- '}NOTIFICADA")

                    case "prepare_correos":
                        self.header("CORREOS")
                        now = date.today()
                        dp = DatePeriod(now.year, now.month)
                        clients = ClientsManager()
                        a_invoices = self.get_all_invoices()

                        cfdi_correos = []
                        for receptor_rfc, notify_invoices in itertools.groupby(
                                sorted(
                                    (i for i in a_invoices.values() if i.estatus == "1" and not self.local_db.notified(i)),
                                    key=lambda r: r["Receptor"]["Rfc"]
                                ),
                                lambda r: r["Receptor"]["Rfc"]
                        ):
                            notify_invoices = list(notify_invoices)

                            def fac_iter():
                                for i in self.get_all_invoices().values():
                                    if i["Emisor"]["Rfc"] == self.csd_signer.rfc \
                                            and self.local_db.liquidated_state(i) == LiquidatedState.NO \
                                            and i["Fecha"] < dp \
                                            and i["Receptor"]["Rfc"] == receptor_rfc:
                                        yield i

                            fac_pen = sorted(
                                fac_iter(),
                                key=lambda r: r["Fecha"]
                            )
                            receptor = clients[receptor_rfc]
                            cfdi_correos.append((receptor, notify_invoices, fac_pen))

                        if cfdi_correos:
                            print(tabulate(
                                [
                                    [
                                        i,
                                        receptor["RazonSocial"][0:36],
                                        receptor["Rfc"],
                                        ",".join(n.name for n in notify_invoices),
                                        ",".join(n.name for n in facturas_pendientes)
                                    ] for i, (receptor, notify_invoices, facturas_pendientes) in enumerate(cfdi_correos, start=1)
                                ],
                                headers=(
                                    '',
                                    'Receptor Razon Social',
                                    'Recep. Rfc',
                                    'Facturas',
                                    'Pendientes Emitidas Meses Anteriores'
                                ),
                                disable_numparse=True,
                                colalign=("right", "left", "left", "left", "left")
                            ))

                            self.action_button_manager.set_items(ACTION_EMAILS, cfdi_correos)
                        else:
                            print("No hay correos pendientes de enviar")

                    case "crear_facturas":
                        self.console.update(autoscroll=True)
                        self.header(f"PROCESAR {action_name.upper()}", clear=False)
                        res = PySimpleGUI.popup(
                            f"Estas seguro que quieres crear {len(action_items)} {action_name}?",
                            title=self.window[event].ButtonText,
                            button_type=POPUP_BUTTONS_OK_CANCEL,
                        )
                        if res == "OK":
                            self.action_button(action_name, action_items)
                            print("FIN")
                        else:
                            print("OPERACION CANCELADA")
                        self.console.update(autoscroll=False)

                    case "facturas_pendientes":
                        self.header("FACTURAS PENDIENTES")

                        def fac_iter():
                            for i in self.get_all_invoices().values():
                                if i["Emisor"]["Rfc"] == self.csd_signer.rfc \
                                        and self.local_db.liquidated_state(i) == LiquidatedState.NO:
                                    yield i

                        self.print_satcfdis(fac_iter())

                    case "descarga":
                        self.header('DESCARGADA')
                        try:
                            res = self.pac_service.recover(values["factura_pagar"], accept=Accept.XML_PDF)
                            self.all_invoices = None
                            cfdi = move_to_folder(res.xml, pdf_data=res.pdf)
                            cfdi = self.get_all_invoices()[to_uuid(cfdi["Complemento"]["TimbreFiscalDigital"]["UUID"])]
                            self.print_satcfdis([cfdi])
                        except DocumentNotFoundError:
                            logger.info("Factura no encontrada")

                    case "ver_excel":
                        self.header("EXCEL")
                        clients = ClientsManager()
                        emisor_cif = clients[self.csd_signer.rfc]
                        exportar_facturas(
                            self.get_all_invoices(),
                            parse_date_period(values["periodo"]),
                            emisor_cif,
                            self.rfc_prediales
                        )

                        archivo_excel = facturas_filename(parse_date_period(values["periodo"]))
                        os.startfile(
                            os.path.abspath(archivo_excel)
                        )

                    case "facturas_emitidas" | "periodo_enter":
                        self.header(f"FACTURAS EMITIDAS {values['periodo']}")
                        dp = parse_date_period(values["periodo"])

                        def fact_iter():
                            for i in self.get_all_invoices().values():
                                if i["Emisor"]["Rfc"] == self.csd_signer.rfc \
                                        and i["Fecha"] == dp:
                                    yield i

                        self.print_satcfdis(fact_iter())

                    case "ver_html":
                        self.header("HTML")
                        dp = parse_date_period(values["periodo"])

                        def fact_iter():
                            for i in self.get_all_invoices().values():
                                if i["Emisor"]["Rfc"] == self.csd_signer.rfc \
                                        and i["Fecha"] == dp:
                                    yield i

                        outfile = facturas_filename(dp, ext="html")
                        Representable.html_write_all(
                            objs=fact_iter(),
                            target=outfile,
                        )
                        os.startfile(
                            os.path.abspath(outfile)
                        )

                    case "ver_carpeta":
                        archivo = facturas_filename(parse_date_period(values["periodo"]))
                        os.startfile(
                            os.path.abspath(os.path.dirname(archivo))
                        )

                    case "ver_carpeta_ajustes":
                        ajustes_dir = ajustes_directory(parse_ym_date(values['periodo']))
                        os.startfile(
                            os.path.abspath(ajustes_dir)
                        )

                    case "periodo" | "inicio" | "final" | "fecha_pago" | "forma_pago" | "importe_pago" | ' ':
                        pass

                    case _:
                        logger.error(f"Unknown event '{event}'")

            except Exception as ex:
                logger.exception(header_line("ERROR"))



import logging
import os

import PySimpleGUI as sg

SOURCE_DIRECTORY = os.path.dirname(__file__)
DATA_DIRECTORY = ".data"
PPD = "PPD"
PUE = "PUE"


def add_file_handler():
    os.makedirs(DATA_DIRECTORY, exist_ok=True)
    fh = logging.FileHandler(
        os.path.join(DATA_DIRECTORY, 'errors.log'),
        mode='a',
        encoding='utf-8',
    )
    fh.setLevel(logging.ERROR)
    formatter = logging.Formatter('%(asctime)s - %(message)s')
    fh.setFormatter(formatter)
    logging.root.addHandler(fh)


class FacturacionLauncher:

    def __init__(self, debug=False):
        # set up logging
        if debug:
            logging.basicConfig(level=logging.ERROR)
        add_file_handler()

        layout = [
            [sg.Column([[
                sg.Image(source=os.path.join(SOURCE_DIRECTORY, "images", "logo.png"), pad=2),
            ]], justification='center', pad=0, background_color='black')],
            [
                sg.Multiline(
                    "Cargando Aplicación...",
                    expand_x=True,
                    expand_y=True,
                    key="console",
                    background_color=sg.theme_background_color(),  # '#a0dbd9',
                    no_scrollbar=True,
                    border_width=0,
                    write_only=True,
                    disabled=True,
                )
            ]
        ]

        self.window = sg.Window(
            f"Facturación Masiva CFDI 4.0",  # {self.csd_signer.rfc}
            layout,
            size=(640, 480),
            resizable=True,
            font=("Courier New", 10, "bold"),
            no_titlebar=True,
            modal=True,
            background_color=sg.theme_background_color(),
            auto_close=True,
            auto_close_duration=6,  # seconds
        )

    @staticmethod
    def read_config():
        from satdigitalinvoice.file_data_managers import ConfigManager
        return ConfigManager()

    def run(self):
        self.window.finalize()
        self.window.read(timeout=0)

        try:
            from satdigitalinvoice.facturacion import FacturacionGUI
            config = self.read_config()
            app = FacturacionGUI(config)
        except Exception as ex:
            logging.exception(ex)
            self.window['console'].update(
                str(ex)
            )
            self.window.read()
            self.window.close()
            return

        self.window.close()
        app.run()



import logging
import os
from datetime import datetime, date
from decimal import Decimal

import xlsxwriter
from babel.dates import format_date
from markdown2 import markdown
from satcfdi import DatePeriod
from satcfdi.accounting import filter_invoices_iter, filter_payments_iter, invoices_export, payments_export, SatCFDI
from satcfdi.create.cfd import cfdi40
from satcfdi.create.cfd.cfdi40 import Comprobante, PagoComprobante
from satcfdi.pacs import sat
# noinspection PyUnresolvedReferences
from satcfdi.transform.catalog import CATALOGS
from tabulate import tabulate
from weasyprint import HTML, CSS
from xlsxwriter.exceptions import FileCreateError

from . import SOURCE_DIRECTORY, PPD
from .environments import facturacion_environment
from .file_data_managers import ClientsManager, FacturasManager
from .formatting_functions.common import fecha, pesos, porcentaje
from .log_tools import print_yaml
from .utils import clear_directory, find_best_match, months_between, add_month

logger = logging.getLogger(__name__)
logger.level = logging.INFO

sat_manager = sat.SAT()

PERIODOS = {
    "Mensual": 1,
    "Bimestral": 2,
    "Trimestral": 3,
    "Cuatrimestral": 4,
    "Semestral": 6,
    "Anual": 12
}


def create_cfdi(receptor_cif, factura_details, serie, folio, csd_sginer, emisor_cif):
    emisor = cfdi40.Emisor(
        rfc=emisor_cif['Rfc'],
        nombre=emisor_cif['RazonSocial'],
        regimen_fiscal=emisor_cif['RegimenFiscal']
    )
    emisor = emisor | factura_details.get('Emisor', {})
    invoice = cfdi40.Comprobante(
        emisor=emisor,
        lugar_expedicion=emisor_cif['CodigoPostal'],
        receptor=cfdi40.Receptor(
            rfc=factura_details['Receptor'],
            nombre=receptor_cif['RazonSocial'],
            uso_cfdi=factura_details['UsoCFDI'],
            domicilio_fiscal_receptor=receptor_cif['CodigoPostal'],
            regimen_fiscal_receptor=receptor_cif['RegimenFiscal']
        ),
        metodo_pago=factura_details['MetodoPago'],
        forma_pago=factura_details['FormaPago'],
        serie=serie,
        folio=folio,
        conceptos=factura_details["Conceptos"]
    )
    invoice.sign(csd_sginer)
    invoice = invoice.process()

    expected_total = factura_details.get('Total')
    if expected_total is not None:
        if expected_total != invoice['Total']:
            logger.info(f"{factura_details['Receptor']}: Total '{expected_total}' is invalid, expected '{invoice['Total']}'")
            return None

    return invoice


def parse_periodo_mes_ajuste(periodo_mes_ajuste: str):
    parts = periodo_mes_ajuste.split(".")
    if len(parts) != 2:
        raise ValueError("Periodo Invalido")

    periodo, mes_ajuste = parts
    if not mes_ajuste.isnumeric():
        raise ValueError("Periodo Invalido")

    mes_ajuste = int(mes_ajuste)
    if not (12 >= int(mes_ajuste) >= 1):
        raise ValueError("Periodo Invalido")

    if periodo not in PERIODOS:
        raise ValueError("Periodo Invalido")

    return periodo, mes_ajuste


def format_concepto_desc(concepto, periodo):
    concepto = concepto.copy()
    template = facturacion_environment.from_string(concepto["Descripcion"])
    concepto["Descripcion"] = template.render(
        periodo=periodo
    )
    return concepto


def validad_facturas(clients, facturas):
    is_valid = True
    for factura_details in facturas:
        cliente = clients.get(factura_details['Receptor'])
        if not cliente:
            logger.info(f"{factura_details['Receptor']}: client not found")
            is_valid = False

        for c in factura_details["Conceptos"]:
            periodo_mes_ajuste = c.get("_periodo_mes_ajuste", "")
            try:
                parse_periodo_mes_ajuste(periodo_mes_ajuste)
            except ValueError:
                logger.info(f"{factura_details['Receptor']}: _periodo_mes_ajuste '{periodo_mes_ajuste}' is invalid")
                is_valid = False

        if factura_details["MetodoPago"] == "PPD" and factura_details["FormaPago"] != "99":
            logger.info(f"{factura_details['Receptor']}: FormaPago '{factura_details['FormaPago']}' is invalid, expected '99' for PPD")
            is_valid = False

    return is_valid


def year_month_desc(year, month):
    return format_date(date(year=year, month=month, day=1), locale='es_MX', format="'Mes de' MMMM 'del' y").upper()


def periodo_desc(ym_date, periodo_mes_ajuste):
    periodo, mes_ajuste = parse_periodo_mes_ajuste(periodo_mes_ajuste)
    periodo_meses = PERIODOS[periodo]

    if (ym_date.month - mes_ajuste) % periodo_meses == 0:
        periodo = year_month_desc(ym_date.year, ym_date.month)
        if periodo_meses > 1:
            mes_final = (ym_date.month + periodo_meses - 2) % 12 + 1
            periodo += " AL " + year_month_desc(
                year=ym_date.year + int(mes_final < ym_date.month),
                month=mes_final
            )

        return periodo
    return None


def generate_ingresos(folio, serie, clients, facturas, ym_date, csd_signer):
    if not validad_facturas(clients, facturas):
        return

    emisor_cif = clients[csd_signer.rfc]

    def prepare_concepto(concepto):
        periodo = periodo_desc(ym_date, concepto['_periodo_mes_ajuste'])
        if periodo and concepto['ValorUnitario'] is not None:
            return format_concepto_desc(concepto, periodo=periodo)

    def facturas_iter():
        i = 0
        for f in facturas:
            receptor_cif = clients[f['Receptor']]
            conceptos = [x for x in (prepare_concepto(c) for c in f["Conceptos"]) if x]
            if conceptos:
                f["Conceptos"] = conceptos
                yield create_cfdi(receptor_cif, f, serie, str(folio + i), csd_signer, emisor_cif)
                i += 1

    cfdis = [i for i in facturas_iter()]

    if None in cfdis:
        return

    return cfdis


def parse_fecha_pago(fecha_pago):
    if not fecha_pago:
        logger.info("Fecha de Pago esta vacia")
        return

    fecha_pago = datetime.fromisoformat(fecha_pago)
    if fecha_pago > datetime.now():
        logger.info("Fecha de Pago esta en el futuro")
        return

    if fecha_pago.replace(hour=12) > datetime.now():
        fecha_pago = datetime.now()
    else:
        fecha_pago = fecha_pago.replace(hour=12)

    dif = datetime.now() - fecha_pago
    if dif.days > 30:
        logger.info("!!! FECHA DE PAGO ES MAYOR A 30 DIAS !!!")

    return fecha_pago


def pago_factura(serie, folio, factura_pagar, fecha_pago, importe_pago, forma_pago, csd_signer):
    if not (fecha_pago := parse_fecha_pago(fecha_pago)):
        return

    c = factura_pagar

    if importe_pago:
        importe_pago = round(Decimal(importe_pago), 2)
    else:
        importe_pago = c.saldo_pendiente

    invoice = Comprobante.pago_comprobantes(
        comprobantes=[
            PagoComprobante(
                comprobante=c,
                num_parcialidad=c.ultima_num_parcialidad + 1,
                imp_saldo_ant=c.saldo_pendiente,
                imp_pagado=importe_pago
            )
        ],
        fecha_pago=fecha_pago,
        forma_pago=forma_pago,
        serie=serie,
        folio=str(folio),
    )
    invoice.sign(csd_signer)
    return invoice.process()


def find_ajustes(facturas, mes_ajuste):
    for f in facturas:
        rfc = f["Receptor"]
        for concepto in f["Conceptos"]:
            _, mes_aj = parse_periodo_mes_ajuste(concepto['_periodo_mes_ajuste'])
            if mes_aj == mes_ajuste:
                yield rfc, concepto


def facturas_folder(dp: DatePeriod):
    if dp.month:
        return f"facturas/{dp.year}/{dp}"
    return f"facturas/{dp.year}/{dp}"


def facturas_filename(dp: DatePeriod, ext="xlsx"):
    return os.path.join(facturas_folder(dp), f"{dp}.{ext}")


def exportar_facturas(all_invoices, dp: DatePeriod, emisor_cif, rfc_prediales):
    emisor_rfc = emisor_cif['Rfc']
    emisor_regimen = emisor_cif['RegimenFiscal']

    emitidas = filter_invoices_iter(invoices=all_invoices.values(), fecha=dp, rfc_emisor=emisor_rfc)
    emitidas_pagos = filter_payments_iter(invoices=all_invoices, fecha=dp, rfc_emisor=emisor_rfc)

    recibidas = filter_invoices_iter(invoices=all_invoices.values(), fecha=dp, rfc_receptor=emisor_rfc)
    recibidas_pagos = filter_payments_iter(invoices=all_invoices, fecha=dp, rfc_receptor=emisor_rfc)

    recibidas_pagos = list(recibidas_pagos)
    pagos_hechos_iva = [
        p
        for p in recibidas_pagos
        if sum(x.get("Importe", 0) for x in p.impuestos.get("Traslados", {}).values()) > 0
           and p.comprobante["Receptor"].get("RegimenFiscalReceptor") in (emisor_regimen, None)
    ]
    prediales = [p for p in recibidas_pagos if p.comprobante["Emisor"]["Rfc"] in rfc_prediales]

    archivo_excel = facturas_filename(dp)
    os.makedirs(os.path.dirname(archivo_excel), exist_ok=True)

    workbook = xlsxwriter.Workbook(archivo_excel)
    # EMITIDAS
    invoices_export(workbook, "EMITIDAS", emitidas)
    payments_export(workbook, "EMITIDAS PAGOS", emitidas_pagos)

    # RECIBIDAS
    invoices_export(workbook, "RECIBIDAS", recibidas)
    payments_export(workbook, "RECIBIDAS PAGOS", recibidas_pagos)

    # SPECIALES
    payments_export(workbook, f"RECIBIDAS PAGOS IVA {emisor_regimen.code}", pagos_hechos_iva)
    if prediales:
        payments_export(workbook, "PREDIALES", prediales)

    try:
        workbook.close()
        print(f"Archivo {archivo_excel} creado")
    except FileCreateError:
        print(f"No se pudo crear el archivo {archivo_excel}")


def generate_pdf_template(template_name, fields):
    increment_template = facturacion_environment.get_template(template_name)
    md5_document = increment_template.render(
        fields
    )
    html = markdown(md5_document)
    pdf = HTML(string=html).write_pdf(
        target=None,
        stylesheets=[
            os.path.join(SOURCE_DIRECTORY, "markdown_styles", "markdown6.css"),
            CSS(
                string='@page { width: Letter; margin: 1.6cm 1.6cm 1.6cm 1.6cm; }'
            )
        ]
    )
    return pdf


def generate_html_template(template_name, fields):
    increment_template = facturacion_environment.get_template(template_name)
    render = increment_template.render(
        fields
    )
    return render


def mf_pago_fmt(cfdi):
    i = cfdi
    if i['TipoDeComprobante'] == "I":
        return i['TipoDeComprobante'].code + ' ' + i['MetodoPago'].code + ' ' + (i['FormaPago'].code if i['FormaPago'].code != '99' else '')
    return i['TipoDeComprobante'].code


def print_invoices(invoices):
    print(
        tabulate(
            invoices,
            headers=(
                '',
                'Receptor Razon Social',
                'Recep. Rfc',
                'Factura',
                "Fecha",
                "Total",
                "Pagada",
                "Tipo",
                "Folio Fiscal",
                "🛈"
            ),
            disable_numparse=True,
            colalign=("right", "left", "left", "left", "left", "right", "right", "left", "left", "left")
        )
    )


def print_cfdis(cfdis, start):
    print(
        tabulate(
            [
                [
                    i,
                    cfdi['Emisor']['RegimenFiscal'].code,
                    cfdi['Receptor']['Nombre'][0:36],
                    cfdi['Receptor']['Rfc'],
                    cfdi.get('Serie', '') + cfdi.get('Folio', ''),
                    mf_pago_fmt(cfdi),
                    cfdi['SubTotal'],
                    cfdi['Total'],
                ]
                for i, cfdi in enumerate(cfdis, start=start)
            ],
            floatfmt=".2f",
            headers=(
                '',
                'EReg',
                'Receptor Razon Social',
                'Recep. Rfc',
                'Factura',
                "Tipo",
                "Subtotal",
                "Total",
            ),
            disable_numparse=True,
            colalign=("right", "left", "left", "left", "left", "left", "right", "right")
        )
    )


def print_cfdi_details(cfdi):
    i = cfdi
    if i["TipoDeComprobante"] == "P":
        pagos = i["Complemento"]["Pagos"]["Pago"]
        print_yaml({
            "Pago": [{
                "FechaPago": p["FechaPago"],
                "FormaDePagoP": p["FormaDePagoP"],
                "Monto": p["Monto"],
                "DoctoRelacionado": [{
                    "Factura": d.get("Serie", "") + d.get("Folio", ""),
                    "IdDocumento": d.get("IdDocumento"),
                    "NumParcialidad": d.get("NumParcialidad"),
                    "ImpSaldoAnt": d.get("ImpSaldoAnt"),
                    "ImpPagado": d.get("ImpPagado"),
                    "ImpSaldoInsoluto": d.get("ImpSaldoInsoluto"),
                } for d in p["DoctoRelacionado"]]
            } for p in pagos]
        })
    else:
        if isinstance(cfdi, SatCFDI) and cfdi.get("MetodoPago") == PPD:
            print_yaml({
                "Conceptos": [x['Descripcion'] for x in i["Conceptos"]],
                "Pendiente": cfdi.saldo_pendiente,
                "Pagos": [f"{c.comprobante.name} - {c.comprobante.uuid}" for c in cfdi.payments if c.comprobante.estatus != '0']
            })
        else:
            print_yaml({
                "Conceptos": [x['Descripcion'] for x in i["Conceptos"]]
            })


def ajustes_directory(ym_date):
    return os.path.join(facturas_folder(DatePeriod(year=ym_date.year, month=ym_date.month)), 'ajustes')


def ajustes(emisor_rfc, ym_date):
    ym_date_effective = add_month(ym_date)

    # clear directory
    ajustes_dir = ajustes_directory(ym_date)
    clear_directory(ajustes_dir)

    clients = ClientsManager()
    facturas = FacturasManager(None)["Facturas"]

    def ajustes_iter():
        for i, (rfc, concepto) in enumerate(find_ajustes(facturas, ym_date_effective.month)):
            receptor = clients[rfc]
            valor_unitario_raw = concepto["ValorUnitario"]

            if isinstance(valor_unitario_raw, dict):
                vud, vu = find_best_match(valor_unitario_raw, ym_date)
                vund, vun = find_best_match(valor_unitario_raw, ym_date_effective)
                meses = months_between(vund, vud)
                ajuste_porcentaje = (vun / vu - 1)
            else:
                vu = valor_unitario_raw
                vun = None
                meses = None
                ajuste_porcentaje = None

            concepto = format_concepto_desc(concepto, periodo="INMUEBLE")
            file_name = os.path.join(ajustes_dir, f'AjusteRenta_{rfc}_{i}.pdf')

            data = {
                "receptor": receptor,
                "emisor": clients[emisor_rfc],
                "concepto": concepto,
                "valor_unitario": pesos(vu),
                "valor_unitario_nuevo": pesos(vun) if vun else "",
                "ajuste_porcentaje": porcentaje(ajuste_porcentaje, 2) if ajuste_porcentaje is not None else "",
                "ajuste_periodo": f"{meses} MESES",
                "ajuste_efectivo_al": fecha(ym_date_effective),
                "periodo": concepto['_periodo_mes_ajuste'].split('.')[0].upper(),
                "fecha_hoy": fecha(date.today()),
                'file_name': file_name
            }
            if ajuste_porcentaje is not None:
                res = generate_pdf_template(
                    template_name='incremento_template.md',
                    fields=data
                )
                with open(file_name, 'wb') as f:
                    f.write(res)
            yield data

    ajustes_iter = list(ajustes_iter())
    print("Ajuste Efectivo Al:", fecha(ym_date_effective))
    if ajustes_iter:
        print(
            tabulate(
                [
                    [
                        i,
                        ajuste["receptor"]["RazonSocial"][:36],
                        ajuste["receptor"]["Rfc"],
                        ajuste["valor_unitario"].split(' ')[0],
                        ajuste["valor_unitario_nuevo"].split(' ')[0],
                        ajuste["ajuste_porcentaje"].split(' ')[0],
                        ajuste["periodo"],
                        ajuste["ajuste_periodo"],
                    ]
                    for i, ajuste in enumerate(ajustes_iter, start=1)
                ],
                headers=(
                    "",
                    "Receptor Razon Social",
                    "Recep. Rfc",
                    "Actual",
                    "Nuevo",
                    "Ajuste %",
                    "Periodo",
                    "Ajuste Periodo",
                ),
                disable_numparse=True,
                colalign=("right", "left", "left", "right", "right", "right", "left", "left"),
            )
        )
    else:
        print("No hay ajustes para este mes")



import os
import random
import shutil
from datetime import datetime
from uuid import UUID

from satcfdi import Signer, DatePeriod


def parse_date_period(periodo):
    fmt, period = try_parsing_date(periodo)
    if fmt == '%Y':
        return DatePeriod(period.year)
    if fmt == '%Y-%m':
        return DatePeriod(period.year, period.month)
    if fmt == '%Y-%m-%d':
        return DatePeriod(period.year, period.month, period.day)


def try_parsing_date(text):
    for fmt in ('%Y', '%Y-%m', '%Y-%m-%d'):
        try:
            return fmt, datetime.strptime(text, fmt)
        except ValueError:
            pass
    raise ValueError('no valid date format found')


def parse_ym_date(periodo):
    return try_parsing_date(periodo)[1]


def to_uuid(s):
    try:
        return UUID(s)
    except ValueError:
        return None


def to_int(s):
    try:
        return int(s)
    except ValueError:
        return None


def random_string():
    chars = "0123456789abcdefghijklmnopqrstuvwxzyABCDEFGHIJKLMNOPQRSTUVWXZY"
    return "".join(random.choice(chars) for _ in range(32))


def convert_ans1_date(ans1_date):
    return datetime.strptime(ans1_date.decode('utf-8'), '%Y%m%d%H%M%SZ')


def cert_info(signer: Signer):
    if signer:
        return {
            "NoCertificado": signer.certificate_number,
            "Tipo": str(signer.type),

            "organizationName": signer.certificate.get_subject().O,
            "x500UniqueIdentifier": signer.certificate.get_subject().x500UniqueIdentifier,
            "serialNumber": signer.certificate.get_subject().serialNumber,
            "organizationUnitName": signer.certificate.get_subject().OU,
            "emailAddress": signer.certificate.get_subject().emailAddress,

            "notAfter": convert_ans1_date(signer.certificate.get_notAfter()),
            "notBefore": convert_ans1_date(signer.certificate.get_notBefore()),
        }


def find_best_match(cases, emission_date):
    fk, fv = (None, None)
    for k, v in cases.items():
        k = parse_ym_date(k)
        if k <= emission_date:
            if fk is None or k > fk:
                fk, fv = k, v
    return fk, fv


def clear_directory(directory):
    shutil.rmtree(directory, ignore_errors=True)
    os.makedirs(directory, exist_ok=True)


# calculate the number of months between two dates
def months_between(d1, d2):
    return (d1.year - d2.year) * 12 + d1.month - d2.month


def add_month(d):
    m = d.month % 12 + 1
    y = d.year + d.month // 12
    return d.replace(year=y, month=m)


def load_certificate(data):
    if 'data' in data:
        return Signer.load_pkcs12(
            **data,
        )
    else:
        return Signer.load(
            **data,
        )


def first_duplicate(seq):
    seen = set()
    for x in seq:
        if x in seen:
            return x
        seen.add(x)
    return None



from datetime import date

import PySimpleGUI as sg
# noinspection PyUnresolvedReferences
from satcfdi.transform.catalog import CATALOGS

from .log_tools import *

FORMA_PAGO = CATALOGS['{http://www.sat.gob.mx/sitio_internet/cfd/catalogos}c_FormaPago']
TEXT_PADDING = ((5, 0), 3)
RTEXT_PADDING = ((0, 0), 3)

# 24 x 24
FOLDER_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAAAqElEQVR4nO3UsQnCYBRF4U/EAcRJtHUJcQddwQlsxcpVLF0hOoY2FlaC+SUQQWI0PmwscuF0j3PhFZc2" \
              "/5QR9siRnjhhid6vBVlFXGWHwS8Ft4aCHLPydozFGybo1BV8kl8wLe/muDbcryIFBwzRxbpB/OAcfV8f2y/lD17y7qcbHIPy2oKoILUFqX1RdCpSgMIVHrsUoHDVznVWM9cpQF46imlp8ye5AyE7C1To4" \
              "/HLAAAAAElFTkSuQmCC"

EXCEL_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAAA7klEQVR4nN2VQQqCUBCGv2pfFyhv0FmC7hGFq1bhrgi6gi1cBK6CdhFtvYdCi6CgrZsXwgTyUNOnLmrgBx31" \
             "/+DNOAP/HkNgCqyBK3AHLsDA1HAELIEjcANUjuYm5hbwKDBVKTnyjZ5/ATugkwWwS5qrAsBHkyyA0yBg1SQgAraiSHueCwgBL3V" \
             "/kFwWwAd6Ir8swJXcSZTEvsIRfQXEwFg6y5LruElAogDoioKma1AF4JvUoPUjctsucqi1qVfQpsY1KPuj+SY1qDsqnLaH3SJvF9Qd1wp4ilethTOT9zea7CLzsivzDPSrmPx2vAFRfA9plcmcVwAAAABJRU5ErkJggg=="

HTML_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAAA/ElEQVR4nN2VOwoCMRCGPz2HB7ARHyfwBN5EsLOxWKzs7Ox0e69gYbOtoK2dZ7BVRwZGWJbsJvuw0IGfDUkmX" \
            "/gzm8A/xgi4AC9AHNLo1gGccxaWFCAC1lUBz0CAVIVICYAYpPVNgJSF+ABtYOHoXzcF6AMdYAmsMuo1ATgZpJTvZQAScEa5ETcAiIsA+vcmDm9XgUpsjUJ7tASnGV1tTL8z26VeKTdgAOwz5esFqA6mKDX2WUR3ebe" \
            "+MXCsAkhMLsAGmNQBZC3C7MHRpxY9rO0FvMyWeUUdfIe8a6BMt9S8rqVAmlv7wZECaa43RjYx78kUh3Su5gxDAL8Vb7wE7yjidnvCAAAAAElFTkSuQmCC"

SEARCH_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAABM0lEQVR4nO3Uuy4FURTG8Z8gqOWcROMFKIiCkgcgUYsXELS8wYn" \
              "+FMILEJd4Bg0R1VGoNCoJnWvjkpE1yeTEsMel80+mmOxvrW/ttffa/PMD+rCMQ1zjCRfYwJgfMo5LvJZ8L2ii+7vJHyLRMWbRjx4MoYH7WN9BR9W25JWvo7NEN4Kb0C1UMVguVF6WPGc6tFfokshhBGVtSaEV" \
              "+qlUg3zbWc9TaIZ+MdXgMQJ6E/WN0K+mGlxEwHCifi/086kGmxGQVfYVNdziGYOpBmMxRNk9H/1El939rSjmQEXyg8sOfKak8u3CVGeDWYls/HcLCVphuoZ93LU9G2eoVzXpiAm9" \
              "+uAdeo62TETyb5uICZ3EElYw13ag9d8w+YpaYaLPMeAPqBd2cvIXBrnJKY7e//6RyBvdy1w1bgRSawAAAABJRU5ErkJggg=="

PDF_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAAA+0lEQVR4nN3VPUrEQBQA4E89gCBio7WI" \
           "+INbeAbxAoI3ESzEysoFd21tbe28QixE7a2t7dxGRwKvWCTrTpJNoQ9ek5k3X8hM5vEfo4cXfCFVZBkbbYDnCQunMeAM" \
           "/abAZyaQmiKpBpACmesSSHWRacA8Tiue92cF7GIN57j4kduzAB4DqfXdc4ACQ3xk7FEjYBk7GHQFrOMQl10BJ3GC7roC3mL8vivgNsYXsYmlyKIt8IojLOAYV1jFQWz6sA1QxOnZwv7YFf6OG1xj1AYYRH9YwVPGT5jqXtejeMuHzMXL2tYNJ/2SZe3U6MXESS0zVWQ5t6zZywH+VnwDCQv1frFQIlYAAAAASUVORK5CYII="

DOWNLOAD_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAAA2ElEQVR4nO2SPQrCQBSEv8r0AYPeQ2xyHw8QC4vUovfJDUQJRistvEBstdUqEphACMvmx6TLwINl3" \
                "+x87L6FUT3rDMQMqEw1AoZ9ojmwAjbAGnAtAFeejc7M6sKXwLsUlAGBBRBUvC9gYQOcZIyAPRA2uEEob6Te0Qb4yuR0mIGj3scGMAUcgCvglfqe9uK2n8BkSLR3L/WLddIHYArcKsPM62H4NZ0AJogp" \
                "/C8AeveLKl/TN6CJas8/ZfBpL19nU5tpZxhm1rK2NsBEkOImWYtKFZ5njKKxfvMPfI7/rbo+AAAAAElFTkSuQmCC"

GRID_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAAAZElEQVR4nGNgGAUkAgEGBgYHCrEAPgtACv5TiB2IsaAAyUUXoJiQywtIsQBZ0QEoJgQcBpUFF5AM" \
            "/gDFBwjgC4PKAgdaB5HDqAW4wGgQ/R80qaiA1oXdf1oV1wK0rnBGAQM6AADMiIAFF68RXgAAAABJRU5ErkJggg="

ABOUT_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAABP0lEQVR4nOWVTU7DMBCFvxuwKmrX0G4o5QLtQboDFXEbwrLkGNDucwB" \
             "+VvQObS5ApCxAIz1LCNmOG68QTxopst882y/jMfwHnAIroAQ2ilJjgxzhkYRa4AvYA6" \
             "+KvcZacYx7FBZADXwC98DUw7GxAmjEXRwjbkkfwHkCfyyu5cy7yCPtxhJOUnckruUcgGGMWMqWs8D8UhE6iZ3iMVYtrTwPoVKEUEjDW123qowL" \
             "+uNSGjche6z8YlhGLHI4hGzaAi8dyVWHReiePPsmNprMXeAdeOprUZWwQB2yaKUfNM1Y4Eoa17EyLTIWeIiVqbOp0aXx4U7hw0S565RWsevRKnYpreJ3sxsniE8kntTsHObaTaN" \
             "/MvNwZvK8ETdZ3GGocnMPjom8KeofD846xZYYBuotJmQ31MK+bSzryfwb+AbymF7gpXVM1QAAAABJRU5ErkJggg=="

BUTTON_COLOR = (sg.theme_background_color(), sg.theme_background_color())


def make_layout(has_fiel):
    # LAYOUT
    button_column = [
        sg.Button("Emitidas", key="facturas_emitidas", border_width=0),
        sg.Text("Año-Mes:", pad=TEXT_PADDING),
        sg.Input(date.today().strftime('%Y-%m'), size=(11, 1), key="periodo"),
        sg.Button(image_data=EXCEL_ICON, key="ver_excel", border_width=0, button_color=BUTTON_COLOR),
        sg.Button(image_data=HTML_ICON, key="ver_html", border_width=0, button_color=BUTTON_COLOR),
        sg.Button(image_data=FOLDER_ICON, key="ver_carpeta", border_width=0, button_color=BUTTON_COLOR),

        sg.VSeparator(),
        sg.Button("Preparar", key="prepare_facturas", border_width=0),
        sg.Text("De La:", pad=TEXT_PADDING),
        sg.Input("1", key="inicio", size=(4, 1)),
        sg.Text("Hasta:", pad=TEXT_PADDING),
        sg.Input("", key="final", size=(4, 1)),

        sg.VSeparator(),
        sg.Button("Ajustes", key="preparar_ajuste_anual", border_width=0),
        sg.Button(image_data=FOLDER_ICON, key="ver_carpeta_ajustes", border_width=0, button_color=BUTTON_COLOR),

        sg.Push(),
        sg.Text("Serie:", pad=TEXT_PADDING),
        sg.Text("", key="serie", pad=TEXT_PADDING),
        sg.Text("Folio:", pad=TEXT_PADDING),
        sg.Input("", key="folio", size=(10, 1), enable_events=True),
    ]
    c_second = [
        sg.Column(
            [
                [
                    sg.Button(image_data=SEARCH_ICON, key="buscar_factura", border_width=0, pad=TEXT_PADDING, button_color=BUTTON_COLOR),
                    sg.Input("", size=(40, 1), key="factura_pagar"),
                    sg.Button(image_data=DOWNLOAD_ICON, key="descarga", border_width=0, disabled=True, button_color=BUTTON_COLOR),
                    sg.Button(image_data=PDF_ICON, key="ver_factura", border_width=0, disabled=True, button_color=BUTTON_COLOR),
                ],
                [
                    sg.Button(" ", border_width=0, visible=True, button_color=BUTTON_COLOR),
                    sg.Button(key="status_sat", border_width=0, visible=False),
                    sg.Button(key="email_notificada", border_width=0, visible=False),
                    sg.Button(key="pendiente_pago", border_width=0, visible=False),
                ]
            ],
            pad=0
        ),
        sg.VSeparator(),
        sg.Column(
            [
                [
                    sg.CalendarButton("FechaPago:", format='%Y-%m-%d', title="FechaPago", no_titlebar=False, target="fecha_pago", pad=TEXT_PADDING, border_width=0,
                                      key="fecha_pago_select", disabled=True),
                    sg.Input("", size=(12, 1), key="fecha_pago", disabled=True),
                    sg.Text("FormaPago:", pad=TEXT_PADDING),
                    sg.Combo([Code(k, v) for k, v in FORMA_PAGO.items()], default_value=Code("03", FORMA_PAGO["03"]), key="forma_pago", size=(34, 1),
                             disabled=True),
                ],
                [
                    sg.Button("Comprobante Pago", key="prepare_pago", border_width=0, disabled=True),
                    sg.Text("       ImpPagado:", pad=TEXT_PADDING),
                    sg.Input("", size=(12, 1), key="importe_pago", disabled=True),
                ]
            ],
            pad=0
        )
    ]

    button_column_third = [
        sg.Button("Correos", key="prepare_correos", border_width=0),
        sg.Button("Pendientes", key="facturas_pendientes", border_width=0),
        sg.Button("Clientes", key="prepare_clientes", border_width=0),
        sg.VSeparator(),
        sg.Text("Recuperar:", pad=TEXT_PADDING, visible=has_fiel),
        sg.Button("Emitidas", key="recuperar_emitidas", border_width=0, visible=has_fiel),
        sg.Button("Recibidas", key="recuperar_recibidas", border_width=0, visible=has_fiel),
        sg.Text("Dias:", pad=TEXT_PADDING, visible=has_fiel),
        sg.Input("40", size=(4, 1), key="recuperar_dias", visible=has_fiel),
    ]

    button_column_low = [
        sg.Button("Procesar ", disabled=True, key="crear_facturas", border_width=0),
        sg.Push(),
        sg.Checkbox("Ver Detallado", default=False, key="detallado"),
        sg.Button(image_data=ABOUT_ICON, key="about", border_width=0, button_color=BUTTON_COLOR),
    ]

    # ----- Full layout -----
    return [
        button_column,
        [sg.HSeparator()],
        c_second,
        [sg.HSeparator()],
        button_column_third,
        [sg.Multiline(expand_x=True, expand_y=True, key="console", write_only=True, autoscroll=False, reroute_stdout=True)],
        button_column_low
    ]


class ActionButtonManager:
    def __init__(self, button):
        self._name = ""
        self._items = []
        self.button = button

    def set_items(self, name, items):
        self._name = name
        self._items = items
        self.style_button()

    def clear(self):
        name = self._name
        items = self._items
        self._name = ""
        self._items = []
        self.style_button()
        return name, items

    def style_button(self):
        self.button.update(
            f"Procesar {len(self._items)} {self._name.capitalize()}" if self._items else f"Procesar {self._name.capitalize()}",
            disabled=len(self._items) == 0
        )



import logging
from datetime import date, datetime
from decimal import Decimal

from satcfdi import Signer

from satdigitalinvoice.file_data_managers import ClientsManager, FacturasManager, ConfigManager
from satdigitalinvoice.gui_functions import generate_ingresos, periodo_desc
from satdigitalinvoice.utils import find_best_match

csd_signer = Signer.load(
    certificate=open('csd/cacx7605101p8.cer', 'rb').read(),
    key=open('csd/cacx7605101p8.key', 'rb').read(),
    password=open('csd/cacx7605101p8.txt', 'rb').read(),
)
ym_date = datetime(year=2023, month=4, day=1)
clients = ClientsManager()


def test_generar_ingresos(caplog):
    caplog.set_level(logging.INFO)
    config = ConfigManager()
    facturas = FacturasManager(ym_date)["Facturas"]

    facturas = generate_ingresos(
        serie=config["serie"],
        folio=1000,
        clients=clients,
        facturas=facturas,
        ym_date=date(year=2023, month=4, day=1),
        csd_signer=csd_signer,
    )

    assert caplog.records == []
    assert len(facturas) == 3
    assert facturas[0]["Total"] == Decimal('18065.66')


def test_generar_ingresos_error(caplog):
    caplog.set_level(logging.INFO)
    config = ConfigManager()
    facturas = FacturasManager(ym_date)["FacturasIncorrectas"]
    print(len(facturas))

    ingresos = generate_ingresos(
        serie=config["serie"],
        folio=1000,
        clients=clients,
        facturas=facturas,
        ym_date=date(year=2023, month=4, day=1),
        csd_signer=csd_signer,
    )

    assert len(caplog.records) == 1
    assert caplog.records[0].message == "ABMG891115PD7: Total '41000.16' is invalid, expected '37019.16'"
    assert ingresos is None


def test_generar_ingresos_error2(caplog):
    caplog.set_level(logging.INFO)
    config = ConfigManager()
    facturas = FacturasManager(ym_date)["FacturasIncorrectas2"]
    print(len(facturas))

    ingresos = generate_ingresos(
        serie=config["serie"],
        folio=1000,
        clients=ClientsManager(),
        facturas=facturas,
        ym_date=date(year=2023, month=4, day=1),
        csd_signer=csd_signer,
    )

    assert len(caplog.records) == 1
    assert caplog.records[0].message == "XXXNOEXISTO: client not found"
    assert ingresos is None


def test_periodo_desc():
    assert periodo_desc(date(year=2021, month=1, day=1), 'Mensual.1') == 'MES DE ENERO DEL 2021'
    assert periodo_desc(date(year=2021, month=1, day=1), 'Bimestral.2') is None
    assert periodo_desc(date(year=2021, month=1, day=1), 'Trimestral.3') is None
    assert periodo_desc(date(year=2021, month=1, day=1), 'Cuatrimestral.4') is None
    assert periodo_desc(date(year=2021, month=1, day=1), 'Semestral.5') is None
    assert periodo_desc(date(year=2021, month=1, day=1), 'Anual.6') is None

    assert periodo_desc(date(year=2021, month=12, day=1), 'Mensual.7') == 'MES DE DICIEMBRE DEL 2021'
    assert periodo_desc(date(year=2021, month=12, day=1), 'Bimestral.8') == 'MES DE DICIEMBRE DEL 2021 AL MES DE ENERO DEL 2022'
    assert periodo_desc(date(year=2021, month=12, day=1), 'Trimestral.9') == 'MES DE DICIEMBRE DEL 2021 AL MES DE FEBRERO DEL 2022'
    assert periodo_desc(date(year=2021, month=12, day=1), 'Cuatrimestral.8') == 'MES DE DICIEMBRE DEL 2021 AL MES DE MARZO DEL 2022'
    assert periodo_desc(date(year=2021, month=12, day=1), 'Semestral.6') == 'MES DE DICIEMBRE DEL 2021 AL MES DE MAYO DEL 2022'
    assert periodo_desc(date(year=2021, month=12, day=1), 'Anual.12') == 'MES DE DICIEMBRE DEL 2021 AL MES DE NOVIEMBRE DEL 2022'


def test_find_best_match():
    casesA = {
        '2022-12': Decimal('20.00'),
        '2023-12': Decimal('30.00'),
        '2024-12': Decimal('40.00'),
        '2025-12': None,
    }
    casesB = {
        '2025-12': None,
        '2024-12': Decimal('40.00'),
        '2023-12': Decimal('30.00'),
        '2022-12': Decimal('20.00'),
    }

    for cases in (casesA, casesB):
        assert find_best_match(cases, datetime(2022, 4, 5))[1] is None
        assert find_best_match(cases, datetime(2023, 4, 5))[1] == Decimal('20.00')
        assert find_best_match(cases, datetime(2024, 4, 5))[1] == Decimal('30.00')
        assert find_best_match(cases, datetime(2025, 4, 5))[1] == Decimal('40.00')
        assert find_best_match(cases, datetime(2026, 4, 5))[1] is None
